import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import javax.vecmath.Point2i;


public class WorldModel {
	/**
	 * Returns a set of all transitions that may occur if 
	 * the given action is performed in the given state.
	 * 
	 * @param state the state in which is the action applied
	 * @param action the action applied
	 * @param map the map of the environment
	 * @return the set of transition, where is transition is assigned a probability, next state and the reward received.
	 */
	public static Set<Transition> getTransitions(Point2i state, Action action, CellContent[][] map) {
		Set<Transition> transitions = new HashSet<Transition>();
		
		if (map[state.x][state.y] == CellContent.EMPTY) {
			
			switch (action) {
				case NORTH:
				{
					float prob = 0.8f;
					Point2i succState = nextExpectedState(state, Action.NORTH, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.EAST, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.WEST, map);
					transitions.add(new Transition(prob, succState));
				}
				break;
				
				case SOUTH:
				{
					float prob = 0.8f;
					Point2i succState = nextExpectedState(state, Action.SOUTH, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.EAST, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.WEST, map);
					transitions.add(new Transition(prob, succState));
				}
				break;
				
				case EAST:
				{
					float prob = 0.8f;
					Point2i succState = nextExpectedState(state, Action.EAST, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.NORTH, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.SOUTH, map);
					transitions.add(new Transition(prob, succState));
				}
				break;
				
				case WEST:
				{
					float prob = 0.8f;
					Point2i succState = nextExpectedState(state, Action.WEST, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.NORTH, map);
					transitions.add(new Transition(prob, succState));
				}
				{
					float prob = 0.1f;
					Point2i succState = nextExpectedState(state, Action.SOUTH, map);
					transitions.add(new Transition(prob, succState));
				}	
				break;
			}
		} else {
			transitions.add(new Transition(1f, state));
		}
		
		return transitions;
	}
	
	public static Point2i nextExpectedState(Point2i state, Action action, CellContent[][] map) {
		Point2i desiredTarget = null;
		Point2i actualTarget = null;

		switch(action) {
			case NORTH:
				desiredTarget = new Point2i(state.x,state.y-1);
				break;

			case SOUTH:
				desiredTarget = new Point2i(state.x,state.y+1);
				break;

			case WEST:
				desiredTarget = new Point2i(state.x-1,state.y);
				break;

			case EAST:
				desiredTarget = new Point2i(state.x+1,state.y);
				break;
			default:
				new RuntimeException();
		}

		if (desiredTarget.x < 0 || desiredTarget.x >= map.length || desiredTarget.y < 0 || desiredTarget.y >= map[0].length ||  map[desiredTarget.x][desiredTarget.y] == CellContent.OBSTACLE) {
			actualTarget = new Point2i(state.x,state.y);
		} else {
			actualTarget = desiredTarget;
		}

		return actualTarget;
	}

	static public Point2i nextSimulatedState(Action action, Point2i pos, CellContent[][] map, Random random) {
		Transition[] transitions = WorldModel.getTransitions(pos, action, map).toArray(new Transition[0]);
		double rand = random.nextDouble();
		int i = -1;
		while (rand >= 0) {
			i++;
			rand -= transitions[i].probability;
		}
		return transitions[i].successorState;
	}
}
